#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <vcclr.h>

#using <System.dll>
#using <System.xml.dll>

using namespace System;
using namespace System::Collections;
using namespace System::IO;
using namespace System::IO::Packaging;
using namespace System::Xml;

static wchar_t* s_relTypeSchemaPart         = L"http://schemas.bentley.com/PrintSet/200801/SchemaPart";
static wchar_t* s_relTypePrintSetPart       = L"http://schemas.bentley.com/PrintSet/200801/PrintSetPart";
static wchar_t* s_relTypeFolderContainsItem = L"http://schemas.bentley.com/PrintSet/200801/FolderContainsItem";

/*---------------------------------------------------------------------------------**//**
* @bsimethod                                    Andrew.Edge                     08/2009
+---------------+---------------+---------------+---------------+---------------+------*/
PackageRelationship^ _getPrintSetPartRelationship (Package^ package)
    {
    PackageRelationshipCollection^ relationships = package->GetRelationshipsByType (gcnew String (s_relTypePrintSetPart));
    if (nullptr == relationships)
        return nullptr;

    for each (PackageRelationship^ relationship in relationships)
        return relationship;

    return nullptr;
    }

/*---------------------------------------------------------------------------------**//**
* @bsimethod                                    Andrew.Edge                     08/2009
+---------------+---------------+---------------+---------------+---------------+------*/
static bool     _processXmlNodeList
(
XmlNodeList^    xmlNodeList,
String^         findStr,
String^         replaceStr,
int&            numStringsModified
)
    {
    IEnumerator^ iter = xmlNodeList->GetEnumerator();
    while (iter->MoveNext())
        {
        XmlNode^ iterNode = dynamic_cast<XmlNode^> (iter->Current);

        if (iterNode->HasChildNodes)
            {
            _processXmlNodeList (iterNode->ChildNodes, findStr, replaceStr, numStringsModified);
            continue;
            }

        XmlNode^ parentNode = iterNode->ParentNode;
        
        if ((parentNode->Name != "Ufs") && (parentNode->Name != "Hfs") && (parentNode->Name != "Moniker"))
            continue;

        String^ innerText = iterNode->InnerText;
        if (String::IsNullOrEmpty (innerText))
            continue;

        if (false == innerText->Contains (findStr))
            continue;

        iterNode->InnerText = innerText->Replace (findStr, replaceStr);
        numStringsModified++;
        }

    return true;
    }

/*---------------------------------------------------------------------------------**//**
* @bsimethod                                    Andrew.Edge                     08/2009
+---------------+---------------+---------------+---------------+---------------+------*/
static bool     _processPart
(
Package^        package,
Uri^            partUri,
String^         findStr,
String^         replaceStr,
int&            numStringsModified
)
    {
    bool        isSuccess = false;
    int         numStringsModifiedThisPart = 0;
    XmlReader^  xmlReader = nullptr;
 
    try
        {
        PackagePart^ part = package->GetPart (partUri);

        Stream^ partStream = part->GetStream();

        xmlReader = XmlReader::Create (partStream);

        XmlDocument^ xmlDocument = gcnew XmlDocument();

        xmlDocument->Load (xmlReader);

        XmlNode^ rootNode = xmlDocument->DocumentElement;

        if (false == _processXmlNodeList (xmlDocument->ChildNodes, findStr, replaceStr, numStringsModifiedThisPart))
            goto cleanup;

        PackageRelationshipCollection^ relationships = part->GetRelationshipsByType (gcnew String (s_relTypeFolderContainsItem));
        if (nullptr != relationships)
            {
            for each (PackageRelationship^ relationship in relationships)
                {
                if (false == _processPart (package, relationship->TargetUri, findStr, replaceStr, numStringsModifiedThisPart))
                    goto cleanup;
                }
            }

        if (numStringsModifiedThisPart > 0)
            {
            Stream^ outStream = gcnew MemoryStream();

            XmlTextWriter^ xmlWriter = gcnew XmlTextWriter (outStream, Text::Encoding::Unicode);

            xmlDocument->PreserveWhitespace = true;
            xmlDocument->Save (xmlWriter);

            xmlReader->Close();
            xmlReader = nullptr;

            partStream->Seek (0, SeekOrigin::Begin);
            outStream->Seek (0, SeekOrigin::Begin);

            const int bufferSize = 4096;
            array<Byte>^ bytes = gcnew array<Byte>(bufferSize);
            int numBytes;
            while ((numBytes = outStream->Read (bytes, 0, bufferSize)) > 0)
                partStream->Write (bytes, 0, numBytes);

            xmlWriter->Close();

            numStringsModified += numStringsModifiedThisPart;
            }

        isSuccess = true;
        }
    catch (Exception^ ex)
        {
        pin_ptr<const wchar_t> exceptionW = PtrToStringChars (ex->Message);
        wprintf (L"- _processPart exception: %s\n", exceptionW);
        }

  cleanup:
    if (xmlReader != nullptr)
        xmlReader->Close();

    return isSuccess;
    }

/*---------------------------------------------------------------------------------**//**
* @bsimethod                                    Andrew.Edge                     08/2009
+---------------+---------------+---------------+---------------+---------------+------*/
static void     _processFile
(
String^         fileHfs,
String^         findStr,
String^         replaceStr,
int&            numSuccess,
int&            numFail,
int&            numStringsModified
)
    {
    FileStream^ psetStream = nullptr;
    ZipPackage^ package = nullptr;

    String^ extension = Path::GetExtension (fileHfs);
    if (0 != String::Compare (extension, ".pset", true))
        return;

    pin_ptr<const wchar_t> fileHfsW = PtrToStringChars (fileHfs);
    wprintf (L"%s\n", fileHfsW);

    try
        {
        psetStream = gcnew FileStream (fileHfs, FileMode::Open, FileAccess::ReadWrite, FileShare::ReadWrite);

        package = (ZipPackage^) Package::Open (psetStream, FileMode::Open, FileAccess::ReadWrite);

        PackageRelationship^ printSetPartRelationship = _getPrintSetPartRelationship (package);
        if (nullptr == printSetPartRelationship)
            {
            wprintf (L"- Unable to locate print set part relationship\n");
            numFail++;
            goto cleanup;
            }

        if (false == _processPart (package, printSetPartRelationship->TargetUri, findStr, replaceStr, numStringsModified))
            {
            numFail++;
            goto cleanup;
            }

        numSuccess++;
        }
    catch (Exception^ ex)
        {
        pin_ptr<const wchar_t> exceptionW = PtrToStringChars (ex->Message);
        wprintf (L"- _processFile exception: %s\n", exceptionW);
        numFail++;
        }

  cleanup:
    if (package != nullptr)
        package->Close();

    if (psetStream != nullptr)
        psetStream->Close();
    }

/*---------------------------------------------------------------------------------**//**
* @bsimethod                                    Andrew.Edge                     08/2009
+---------------+---------------+---------------+---------------+---------------+------*/
static void     _processDir
(
const wchar_t*  baseDir,
const wchar_t*  findStr,
const wchar_t*  replaceStr,
const wchar_t*  currentDir,
int&            numSuccess,
int&            numFail,
int&            numStringsModified
)
{
    wchar_t fileNamePattern[MAX_PATH];
    swprintf (fileNamePattern, MAX_PATH, L"%s\\*", currentDir);

    WIN32_FIND_DATA findData;
    HANDLE findFileHandle = FindFirstFile (fileNamePattern, &findData);
    if (findFileHandle != INVALID_HANDLE_VALUE)
        {
        BOOL findMoreFiles = TRUE;
        while (findMoreFiles)
            {
            // Filter out special directories and system files.
            if ((0 != wcscmp (findData.cFileName, L".")) &&
                (0 != wcscmp (findData.cFileName, L"..")) &&
                (! (findData.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM)))
                {
                wchar_t currentFile[MAX_PATH];
                swprintf (currentFile, MAX_PATH, L"%s\\%s", currentDir, findData.cFileName);

                if (findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
                    {
                    _processDir (baseDir, findStr, replaceStr, currentFile, numSuccess, numFail, numStringsModified);
                    }
                else
                    {
                    _processFile (gcnew String (currentFile), gcnew String (findStr), gcnew String (replaceStr),
                                  numSuccess, numFail, numStringsModified);
                    }
                }

            findMoreFiles = FindNextFile (findFileHandle, &findData);
            }

        FindClose (findFileHandle);
        }
    }

/*---------------------------------------------------------------------------------**//**
* @bsimethod                                    Andrew.Edge                     08/2009
+---------------+---------------+---------------+---------------+---------------+------*/
int             wmain (int argc, wchar_t** argv)
    {
    if (argc != 4)
        {
        fwprintf (stderr, L"usage: RebasePset <dir> <findStr> <replaceStr>\n");
        return 1;
        }

    wchar_t baseDir[MAX_PATH];
    wcscpy (baseDir, argv[1]);

    wchar_t findStr[MAX_PATH];
    wcscpy (findStr, argv[2]);

    wchar_t replaceStr[MAX_PATH];
    wcscpy (replaceStr, argv[3]);

    // Make sure directory exists.
    if (-1 == GetFileAttributes (baseDir))
        {
        fwprintf (stderr, L"RebasePset: directory \"%s\" does not exist\n", baseDir);
        return 1;
        }

    int numSuccess = 0;
    int numFail = 0;
    int numStringsModified = 0;

    _processDir (baseDir, findStr, replaceStr, baseDir, numSuccess, numFail, numStringsModified);
    
    if (0 == numFail)
        {
        wprintf (L"RebasePset: %d files processed (%d strings modified)\n", numSuccess, numStringsModified);
        }
    else if (0 == numSuccess)
        {
        wprintf (L"RebasePset: %d files failed processing\n", numFail);
        }
    else
        {
        wprintf (L"RebasePset: %d files processed (%d strings modified), %d files failed processing\n",
                 numSuccess, numStringsModified, numFail);
        }
            
    return 0;
    }
